# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the TU Wien, Department of Geodesy and
#      Geoinformation nor the names of its contributors may be used to endorse
#      or promote products derived from this software without specific prior
#      written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Module containing adapters that can be used together with the validation
framework.
"""

import operator

import pandas as pd
from pytesmo.time_series.anomaly import calc_anomaly
from pytesmo.time_series.anomaly import calc_climatology
from pytesmo.utils import deprecated
from pandas import DataFrame
import numpy as np
import warnings

_op_lookup = {
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    ">": operator.gt,
}


class BasicAdapter:
    """
    Adapter to modify the return value of reading functions from base class.
    - Pick data frame from objects that have a `data_property_name`,
      i.e. ascat time series objects.
    - Removes unnecessary timezone information in pandas data frames which
      pytesmo can not use.
    - adds a method with the name given in `read_name` that calls the same
      method from cls but modifies the returned data frame.
    """

    def __init__(self, cls, data_property_name="data", read_name=None):
        """
        Parameters
        ----------
        cls: object
            The original reader to adapt.
        data_property_name: str, optional (default: "data")
            Attribute name under which the pandas DataFrame containing the time
            series is found in the object returned by the read function of the
            original reader. Ignored if no attribute of this name is found.
            Then it is required that the DataFrame is already the return value
            of the read function.
        read_name: str, optional (default: None)
            To enable the adapter for a method other than `read` or `read_ts`
            give the function name here (a function of that name must exist in
            cls). A method of the same name will be added to the adapted
            Reader, which takes the same arguments as the base method.
            The output of this method will be changed by the adapter.
            If None is passed, only data from `read` and `read_ts` of cls
            will be adapted.
        """

        self.cls = cls
        self.data_property_name = data_property_name
        self.read_name = read_name

        if read_name:
            setattr(self, read_name, self._adapt_custom)

    def __get_dataframe(self, data):
        if ((not isinstance(data, DataFrame)) and
            (hasattr(data, self.data_property_name)) and
            (isinstance(getattr(data, self.data_property_name), DataFrame))):
            data = getattr(data, self.data_property_name)
        return data

    def __drop_tz_info(self, data):
        if hasattr(data.index, "tz") and (data.index.tz is not None):
            warnings.warn(
                f"Dropping timezone information ({data.index.tz})"
                f" for data from reader {self.cls.__class__.__name__}")
            data.index = data.index.tz_convert(None)
        return data

    def _adapt(self, df: DataFrame) -> DataFrame:
        # drop time zone info and extract df from ASCAT TS object
        return self.__drop_tz_info(
            self.__get_dataframe(df) if df is not None else DataFrame())

    def _adapt_custom(self, *args, **kwargs):
        # modifies data from whatever function was set as `read_name`.
        data = getattr(self.cls, self.read_name)(*args, **kwargs)
        return self._adapt(data)

    def read_ts(self, *args, **kwargs):
        data = getattr(self.cls, "read_ts")(*args, **kwargs)
        return self._adapt(data)

    def read(self, *args, **kwargs):
        data = getattr(self.cls, "read")(*args, **kwargs)
        return self._adapt(data)

    @property
    def grid(self):
        """
        Returns grid of wrapped class if it exists, otherwise None.
        """
        if hasattr(self.cls, "grid"):
            return self.cls.grid


@deprecated("`MaskingAdapter` is deprecated, use `SelfMaskingAdapter` "
            "or `AdvancedMaskingAdapter` instead.")
class MaskingAdapter(BasicAdapter):
    """
    Transform the given class to return a boolean dataset given the operator
    and threshold. This class calls the read_ts and read methods
    of the given instance and applies boolean masking to the returned data
    using the given operator and threshold.

    Parameters
    ----------
    cls: object
        Reader object, has to have a `read_ts` or `read` method or a method
        name must be specified in the `read_name` kwarg. The same method will
        be available for the adapted version of the reader.
    op: str or Callable
        Either a string to look up a function from
        :const:`pytesmo/validation_framework/adapters.py._op_lookup`
        or a function that takes `data` and `threshold` as arguments.
    threshold: Any
        Value to use as the threshold combined with the operator to mask
        elements in `column_name`
    column_name: str, optional (default: None)
        Name of the column to apply `op` to. If None is passed,
        nothing happens.
    data_property_name: str, optional (default: "data")
        Attribute name under which the pandas DataFrame containing the time
        series is found in the object returned by the read function of the
        original reader. Ignored if no attribute of this name is found.
        Then it is required that the DataFrame is already the return value
        of the read function.
    read_name: str, optional (default: None)
        To enable the adapter for a method other than `read` or `read_ts`
        give the function name here (a function of that name must exist in
        cls). A method of the same name will be added to the adapted
        Reader, which takes the same arguments as the base method.
        The output of this method will be changed by the adapter.
        If None is passed, only data from `read` and `read_ts` of cls
        will be adapted.
    """

    def __init__(self, cls, op, threshold, column_name=None, **kwargs):

        super().__init__(cls, **kwargs)

        if callable(op):
            self.operator = op
        elif op in _op_lookup:
            self.operator = _op_lookup[op]
        else:
            raise ValueError('"{}" is not a valid operator'.format(op))

        self.threshold = threshold

        self.column_name = column_name

    def _adapt(self, data):
        data = super()._adapt(data)
        if self.column_name is not None:
            data = data.loc[:, [self.column_name]]
        return self.operator(data, self.threshold)


class SelfMaskingAdapter(BasicAdapter):
    """
    Transform the given (reader) class to return a dataset that is masked based
    on the given column, operator, and threshold. This class calls the read_ts
    or read method of the given reader instance, applies the operator/threshold
    to the specified column, and masks the whole dataframe with the result.

    Parameters
    ----------
    cls: object
        Reader object, has to have a `read_ts` or `read` method or a method
        name must be specified in the `read_name` kwarg. The same method will
        be available for the adapted version of the reader.
    op: str or Callable
        Either a string to look up a function from
        :const:`pytesmo/validation_framework/adapters.py._op_lookup`
        or a function that takes `data` and `threshold` as arguments.
    threshold: Any
        Value to use as the threshold combined with the operator to mask
        elements in `column_name`
    column_name: str
        Name of the column to apply `op` to
    data_property_name: str, optional (default: "data")
        Attribute name under which the pandas DataFrame containing the time
        series is found in the object returned by the read function of the
        original reader. Ignored if no attribute of this name is found.
        Then it is required that the DataFrame is already the return value
        of the read function.
    read_name: str, optional (default: None)
        To enable the adapter for a method other than `read` or `read_ts`
        give the function name here (a function of that name must exist in
        cls). A method of the same name will be added to the adapted
        Reader, which takes the same arguments as the base method.
        The output of this method will be changed by the adapter.
        If None is passed, only data from `read` and `read_ts` of cls
        will be adapted.
    """

    def __init__(self, cls, op, threshold, column_name, **kwargs):

        super().__init__(cls, **kwargs)

        if callable(op):
            self.operator = op
        elif op in _op_lookup:
            self.operator = _op_lookup[op]
        else:
            raise ValueError(f"'{op}' is not a valid operator")

        self.threshold = threshold
        self.column_name = column_name

    def _adapt(self, data):
        data = super()._adapt(data)
        mask = self.operator(data[self.column_name], self.threshold)
        return data[mask]


class AdvancedMaskingAdapter(BasicAdapter):
    """
    Transform the given (reader) class to return a dataset that is masked based
    on the given list of filters. A filter is a 3-tuple of column_name,
    operator, and threshold.
    This class calls the reading method of the given reader instance,
    applies all filters separately, ANDs all filters together, and masks the
    whole dataframe with the result.

    Parameters
    ----------
    cls: object
        Reader object, has to have a `read_ts` or `read` method or a method
        name must be specified in the `read_name` kwarg. The same method will
        be available for the adapted version of the reader.
    filter_list: list[tuple, tuple, tuple]
        [column_name, operator, threshold]
        'column_name': string
            name of the column to apply the operator to
        'operator': Callable or str;
            string needs to be one of '<', '<=', '==', '>=', '>', '!=' or a
            function that takes data and threshold as arguments.
        'threshold':
            value to use as the threshold combined with the operator;
    data_property_name: str, optional (default: "data")
        Attribute name under which the pandas DataFrame containing the time
        series is found in the object returned by the read function of the
        original reader. Ignored if no attribute of this name is found.
        Then it is required that the DataFrame is already the return value
        of the read function.
    read_name: str, optional (default: None)
        To enable the adapter for a method other than `read` or `read_ts`
        give the function name here (a function of that name must exist in
        cls). A method of the same name will be added to the adapted
        Reader, which takes the same arguments as the base method.
        The output of this method will be changed by the adapter.
        If None is passed, only data from `read` and `read_ts` of cls
        will be adapted.
    ignore_nans: bool, optional (default: False)
        Should be set to True in case the NaNs in the mask field(s) should
        be ignored, i.e. the main field should not be masked when NaNs
        are present elswehere in the row
    """

    def __init__(self, cls, filter_list, ignore_nans: bool = False, **kwargs):

        super().__init__(cls, **kwargs)

        self.filter_list = filter_list
        self.ignore_nans = ignore_nans

    def _adapt(self, data):
        data = super()._adapt(data)
        mask = None
        for column_name, op, threshold in self.filter_list:
            if callable(op):
                operator = op
            elif op in _op_lookup:
                operator = _op_lookup[op]
            else:
                raise ValueError('"{}" is not a valid operator'.format(op))

            if self.ignore_nans:
                new_mask = operator(data[column_name], threshold) | np.isnan(
                    data[column_name])
            else:
                new_mask = operator(data[column_name], threshold)

            if mask is not None:
                mask = mask & new_mask
            else:
                mask = new_mask

        return data[mask].dropna(how="all")


class AnomalyAdapter(BasicAdapter):
    """
    Takes the pandas DataFrame that reader returns and calculates the
    anomaly of the time series based on a moving average.

    Parameters
    ----------
    cls: object
        Reader object, has to have a `read_ts` or `read` method or a method
        name must be specified in the `read_name` kwarg. The same method will
        be available for the adapted version of the reader.
    window_size : float, optional (default: 35)
        The window-size [days] of the moving-average window to calculate the
        anomaly reference.
    columns: list, optional
        columns in the dataset for which to calculate anomalies.
    data_property_name: str, optional (default: "data")
        Attribute name under which the pandas DataFrame containing the time
        series is found in the object returned by the read function of the
        original reader. Ignored if no attribute of this name is found.
        Then it is required that the DataFrame is already the return value
        of the read function.
    read_name: str, optional (default: None)
        To enable the adapter for a method other than `read` or `read_ts`
        give the function name here (a function of that name must exist in
        cls). A method of the same name will be added to the adapted
        Reader, which takes the same arguments as the base method.
        The output of this method will be changed by the adapter.
        If None is passed, only data from `read` and `read_ts` of cls
        will be adapted.
    """

    def __init__(self, cls, window_size=35, columns=None, **kwargs):

        super().__init__(cls, **kwargs)

        self.window_size = window_size
        self.columns = columns

    def _adapt(self, data):
        data = super()._adapt(data)
        if self.columns is None:
            ite = data
        else:
            ite = self.columns
        for column in ite:
            data[column] = calc_anomaly(
                data[column], window_size=self.window_size)
        return data


class AnomalyClimAdapter(BasicAdapter):
    """
    Takes the pandas DataFrame that reader returns and calculates the
    anomaly of the time series based on the (long-term) average of the series.
    Parameters
    ----------
    cls: object
        Reader object, has to have a `read_ts` or `read` method or a method
        name must be specified in the `read_name` kwarg. The same method will
        be available for the adapted version of the reader.
    columns: list, optional (default: None)
        Columns in the dataset for which to calculate anomalies. If None is
        passed, the anomaly is calculated for all columns.
    data_property_name: str, optional (default: "data")
        Attribute name under which the pandas DataFrame containing the time
        series is found in the object returned by the read function of the
        original reader. Ignored if no attribute of this name is found.
        Then it is required that the DataFrame is already the return value
        of the read function.
    read_name: str, optional (default: None)
        To enable the adapter for a method other than `read` or `read_ts`
        give the function name here (a function of that name must exist in
        cls). A method of the same name will be added to the adapted
        Reader, which takes the same arguments as the base method.
        The output of this method will be changed by the adapter.
        If None is passed, only data from `read` and `read_ts` of cls
        will be adapted.
    return_clim: bool, optional (default: False)
        If True, then a column for the climatology is added to the DataFrame
        returned by the read function.
    kwargs:
        Any remaining keyword arguments will be given to
        :func:`pytesmo.time_series.anomaly.calc_climatology`
    """

    def __init__(self, cls, columns=None, return_clim=False, **kwargs):

        cls_kwargs = dict()
        if "data_property_name" in kwargs:
            cls_kwargs["data_property_name"] = kwargs.pop("data_property_name")
        if "read_name" in kwargs:
            cls_kwargs["read_name"] = kwargs.pop("read_name")

        super().__init__(cls, **cls_kwargs)

        self.return_clim = return_clim
        self.kwargs = kwargs
        self.columns = columns

    def _adapt(self, data):
        data = super()._adapt(data)
        if self.columns is None:
            ite = data
        else:
            ite = self.columns
        for column in ite:
            clim = calc_climatology(data[column], **self.kwargs)
            anom = calc_anomaly(
                data[column], climatology=clim, return_clim=self.return_clim)
            if self.return_clim:
                data[column] = anom['anomaly']
                data[f"{column}_climatology"] = anom['climatology']
            else:
                data[column] = anom

        return data


class ColumnCombineAdapter(BasicAdapter):
    """
    Takes the pandas DataFrame that the read_ts or read method of the instance
    returns and applies a function to merge multiple columns into one.
    E.g. when there are 2 Soil Moisture parameters in a dataset that should be
    averaged on reading. Will add one additional column to the input data
    frame.
    """

    def __init__(
        self,
        cls,
        func,
        func_kwargs=None,
        columns=None,
        new_name="merged",
        **kwargs,
    ):
        """
        Parameters
        ----------
        cls : object
            Reader object, has to have a `read_ts` or `read` method or a
            method name must be specified in the `read_name` kwarg.
            The same method will be available for the adapted version of the
            reader.
        func: Callable
            Will be applied to dataframe columns using
            pd.DataFrame.apply(..., axis=1)
            additional kwargs for this must be given in func_kwargs,
            e.g. :func:`pd.DataFrame.mean`
        func_kwargs : dict, optional (default: None)
            kwargs that are passed to method or None to use the default ones.
        columns: list, optional (default: None)
            Columns in the dataset that are combined. If None are selected
            all columns are used.
        new_name: str, optional (default: merged)
            Name that the merged column will have in the returned data frame.
        data_property_name: str, optional (default: "data")
            Attribute name under which the pandas DataFrame containing the
            time series is found in the object returned by the read function
            of the original reader.
            Ignored if no attribute of this name is found.
            Then it is required that the DataFrame is already the return value
            of the read function.
        read_name: str, optional (default: None)
            To enable the adapter for a method other than `read` or `read_ts`
            give the function name here (a function of that name must exist in
            cls). A method of the same name will be added to the adapted
            Reader, which takes the same arguments as the base method.
            The output of this method will be changed by the adapter.
            If None is passed, only data from `read` and `read_ts` of cls
            will be adapted.
        """

        super().__init__(cls, **kwargs)
        self.func = func
        self.func_kwargs = func_kwargs if func_kwargs is not None else {}
        self.func_kwargs["axis"] = 1
        self.columns = columns
        self.new_name = new_name

    def _adapt(self, data: DataFrame) -> DataFrame:
        data = super()._adapt(data)

        # if DataFrame is empty, needs to be returned in expected format
        if data.empty:
            data[self.new_name] = None
            return data

        columns = data.columns if self.columns is None else self.columns
        new_col = data[columns].apply(self.func, **self.func_kwargs)
        data[self.new_name] = new_col
        return data


class TimestampAdapter(BasicAdapter):
    """
    Class that combines two or more timestamp fields to a single exact
    measurement time. The fields of interest specify:

    1. A basic observation time (e.g. days at midnight) which can
        be expressed in timestamp (YYYY-mm-dd) or with respect to a
        reference time (days since YYYY-mm-dd)
    2. One or more (minute, s, µs) offset times to be added cumulatively

    -------------------
    Example input:

         variable    base_time [w.r.t. 2005-02-01]  offset [min]  offset [sec]
    100  0.889751                            100.0          38.0         999.0
    101  0.108279                            101.0          40.0        1000.0
    102 -1.201708                            102.0          39.0         999.0

    Example output:

                         variable
    2005-05-12 00:55:42  0.889751
    2005-05-13 00:57:39  0.108279
    2005-05-14 00:56:38 -1.201708

    Parameters
    ----------
     cls: object
        Reader object, has to have a `read_ts` or `read` method or a method
        name must be specified in the `read_name` kwarg. The same method will
        be available for the adapted version of the reader.
    time_offset_fields: str, list or None
        name or list of names of the fields that provide information on the time offset.
        If a list is given, all values will contribute to the offset, assuming that
        each refers to the previous. For instance:
        offset = minutes + seconds in the minute + µs in the second
        NOTE: np.nan values are counted as 0 offset
        NOTE: if None, no offset is considered
    time_units: str or list
        time units that the time_offset_fields are specified in. If a list is given,
        it should have the same size as the 'time_offset_fields' parameter
        Can be any of the np.datetime[64] units:
        https://numpy.org/doc/stable/reference/arrays.datetime.html
    base_time_field: str, optional. Default is None.
        If a name is provided, the generic time field will be searched for
        in the columns; otherwise, it is assumed to be the index
        NOTE: np.nan values in this field are dropped
    base_time_reference: str, optional. Default is None.
        String of format 'YYYY-mm-dd' that can be specified to tranform the
        'base_time_field' from [units since base_time_reference] to
        np.datetime[64]. If not provided, it will be assumed that the base_time_field
        is already in np.datetime[64] units
    base_time_units: str, optional. Default is "D"
        Units that the base_time_field is specified in. Only applicable with 'base_time_reference'
    replace_index: bool, optional. Default is True.
        If True, the exact timestamp is used as index. Else, it will be added
        to the dataframe on the column 'output_field'
    output_field: str, optional. Default is None.
        If a name is specified, an additional column is generated under the name,
        with the exact timestamp. Only with 'replace_index' == False
    drop_original: bool, optional. Default is True.
        Whether the base_time_field and time_offset_fields should be dropped in the
        final DataFrame
    """

    def __init__(self,
                 cls: object,
                 time_offset_fields: str or list,
                 time_units: str or list = "s",
                 base_time_field: str = None,
                 base_time_reference: str = None,
                 base_time_units: str = "D",
                 replace_index: bool = True,
                 output_field: str = None,
                 drop_original: bool = True,
                 **kwargs):
        super().__init__(cls, **kwargs)

        if time_offset_fields is None or isinstance(time_offset_fields, list):
            self.time_offset_fields = time_offset_fields
        else:
            self.time_offset_fields = [time_offset_fields]
        self.time_units = time_units if isinstance(time_units,
                                                   list) else [time_units]

        self.base_time_field = base_time_field
        self.base_time_reference = np.datetime64(
            base_time_reference) if base_time_reference is not None else None
        self.base_time_units = base_time_units

        self.replace_index = replace_index
        if not replace_index and output_field is None:
            raise ValueError(
                "'output_field' should be specified in case the new timestamp"
                "should not be used as index. Alternatively, set 'replace_index' to True"
            )
        elif replace_index and output_field is not None:
            warnings.warn(
                "Ignoring the 'output_field' value. Set 'replace_index' to True to"
                "avoid this behavior")
        else:
            self.output_field = output_field

        self.drop_original = drop_original

    def convert_generic(self,
                        time_arr: np.array,
                        units: str = 'D') -> np.array:
        """Convert the generic time field to np.datetime[64] dtype"""
        time_delta = time_arr.astype(int).astype(f'timedelta64[{units}]')
        time_date = np.full(time_delta.shape,
                            self.base_time_reference) + time_delta

        return time_date

    def add_offset_cumulative(self, data: DataFrame) -> np.array:
        """Return an array of timedelta calculated with all the time_offset_fields"""
        total_offset = np.full(data.index.shape, 0, dtype='timedelta64[s]')
        for field, unit in zip(self.time_offset_fields, self.time_units):
            total_offset += data[field].map(
                lambda x: np.timedelta64(int(x), unit)
                if not np.isnan(x) else np.timedelta64(0, unit)).values

        return total_offset

    def _adapt(self, data: DataFrame) -> DataFrame:
        """
        Adapt the timestamps in the original with the specified offset
        NOTE: assumes the index dtype is 'datetime64[ns]'
        """
        data = super()._adapt(data)
        original = data.copy()

        # Get the generic time array
        if self.base_time_field is not None:
            base_time = data[self.base_time_field]
        else:
            base_time = data.index

        # Take only the valid dates
        data = data[base_time.notna()]
        base_time_values = base_time.dropna().values

        # Make sure the dataframes contains values after dropna()
        if data.empty:
            warnings.warn(
                "The input DataFrame is either empty or has all NaT values"
                " in the specified `base_time_field`, therefore the original"
                " non-adapted is returned")

            return original

        if self.base_time_reference is not None:
            base_time_values = self.convert_generic(base_time_values,
                                                    self.base_time_units)

        # If no offset is specified
        if self.time_offset_fields is None:
            exact_time = base_time_values
        else:
            # Add time offset to the generic time
            offset = self.add_offset_cumulative(data)
            exact_time = base_time_values + offset

        # generate the final frame
        if not self.replace_index:
            data[self.output_field] = exact_time
        else:
            data.index = exact_time

        if self.drop_original:
            if self.time_offset_fields is not None:
                data.drop(columns=self.time_offset_fields, inplace=True)
            if self.base_time_field in data.columns:
                data.drop(columns=[self.base_time_field], inplace=True)

        # Remove NaNs from index, if present
        data = data.loc[data.index.dropna()]

        return data



# Copyright (c) 2015,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology,
#      Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Tests for the temporal matching module
Created on Wed Jul  8 19:37:14 2015
"""


from copy import deepcopy
from datetime import datetime
import numpy as np
import numpy.testing as nptest
import pandas as pd
import pytesmo.temporal_matching as tmatching
import pytest


def test_df_match_borders():
    """
    Border values can be problematic for temporal matching.

    See issue #51
    """

    ref_df = pd.DataFrame(
        {"data": np.arange(5)},
        index=pd.date_range(datetime(2007, 1, 1, 0), "2007-01-05", freq="D"),
    )
    match_df = pd.DataFrame(
        {"matched_data": np.arange(5)},
        index=[
            datetime(2007, 1, 1, 9),
            datetime(2007, 1, 2, 9),
            datetime(2007, 1, 3, 9),
            datetime(2007, 1, 4, 9),
            datetime(2007, 1, 5, 9),
        ],
    )
    with pytest.deprecated_call():
        matched = tmatching.df_match(ref_df, match_df)

    nptest.assert_allclose(
        np.array([0.375, 0.375, 0.375, 0.375, 0.375]), matched.distance.values
    )
    nptest.assert_allclose(np.arange(5), matched.matched_data)


def test_df_match_match_on_window_border():
    """
    test matching if a value lies exactly on the window border.
    """

    ref_df = pd.DataFrame(
        {"data": np.arange(5)},
        index=pd.date_range(datetime(2007, 1, 1, 0), "2007-01-05", freq="D"),
    )
    match_df = pd.DataFrame(
        {"matched_data": np.arange(4)},
        index=[
            datetime(2007, 1, 1, 9),
            datetime(2007, 1, 2, 9),
            datetime(2007, 1, 3, 12),
            datetime(2007, 1, 5, 9),
        ],
    )
    with pytest.deprecated_call():
        matched = tmatching.df_match(ref_df, match_df, window=0.5)

    nptest.assert_allclose(
        np.array([0.375, 0.375, 0.5, -0.5, 0.375]), matched.distance.values
    )
    nptest.assert_allclose([0, 1, 2, 2, 3], matched.matched_data)

    # test asym_window keyword
    with pytest.deprecated_call():
        matched = tmatching.df_match(
            ref_df, match_df, window=0.5, asym_window="<="
        )

    nptest.assert_allclose(
        np.array([0.375, 0.375, 0.5, np.nan, 0.375]), matched.distance.values
    )
    nptest.assert_allclose([0, 1, 2, np.nan, 3], matched.matched_data)

    with pytest.deprecated_call():
        matched = tmatching.df_match(
            ref_df, match_df, window=0.5, asym_window=">="
        )

    nptest.assert_allclose(
        np.array([0.375, 0.375, np.nan, -0.5, 0.375]), matched.distance.values
    )
    nptest.assert_allclose([0, 1, np.nan, 2, 3], matched.matched_data)


def test_df_match_borders_unequal_query_points():
    """
    Border values can be problematic for temporal matching.

    See issue #51
    """

    ref_df = pd.DataFrame(
        {"data": np.arange(5)},
        index=pd.date_range(datetime(2007, 1, 1, 0), "2007-01-05", freq="D"),
    )
    match_df = pd.DataFrame(
        {"matched_data": np.arange(4)},
        index=[
            datetime(2007, 1, 1, 9),
            datetime(2007, 1, 2, 9),
            datetime(2007, 1, 4, 9),
            datetime(2007, 1, 5, 9),
        ],
    )
    with pytest.deprecated_call():
        matched = tmatching.df_match(ref_df, match_df)

    nptest.assert_allclose(
        np.array([0.375, 0.375, -0.625, 0.375, 0.375]), matched.distance.values
    )
    nptest.assert_allclose(np.array([0, 1, 1, 2, 3]), matched.matched_data)


def test_matching():
    """
    test matching function
    """
    data = np.arange(5.0)
    data[3] = np.nan

    ref_df = pd.DataFrame(
        {"data": data},
        index=pd.date_range(datetime(2007, 1, 1, 0), "2007-01-05", freq="D"),
    )
    match_df = pd.DataFrame(
        {"matched_data": np.arange(5)},
        index=[
            datetime(2007, 1, 1, 9),
            datetime(2007, 1, 2, 9),
            datetime(2007, 1, 3, 9),
            datetime(2007, 1, 4, 9),
            datetime(2007, 1, 5, 9),
        ],
    )
    with pytest.deprecated_call():
        matched = tmatching.matching(ref_df, match_df)

    nptest.assert_allclose(np.array([0, 1, 2, 4]), matched.matched_data)
    assert len(matched) == 4


def test_matching_series():
    """
    test matching function with pd.Series as input
    """
    data = np.arange(5.0)
    data[3] = np.nan

    ref_ser = pd.Series(
        data,
        index=pd.date_range(datetime(2007, 1, 1, 0), "2007-01-05", freq="D"),
    )
    match_ser = pd.Series(
        np.arange(5),
        index=[
            datetime(2007, 1, 1, 9),
            datetime(2007, 1, 2, 9),
            datetime(2007, 1, 3, 9),
            datetime(2007, 1, 4, 9),
            datetime(2007, 1, 5, 9),
        ],
        name="matched_data",
    )

    with pytest.deprecated_call():
        matched = tmatching.matching(ref_ser, match_ser)

    nptest.assert_allclose(np.array([0, 1, 2, 4]), matched.matched_data)
    assert len(matched) == 4


#############################################################################
# Tests for new implementation
#############################################################################


@pytest.fixture
def test_data():
    """
    Test data for temporal matching.

    The test frames have modified time indices:
    - shifted by 3 hours
    - shifted by 7 hours
    - shifted by 3 hours, and in a timezone such that displayed numbers are >6
      hours apart
    - shifted by 7 hours, and in a timezone such that displayed numbers are <6
      hours apart
    - randomly shifted by a value between -12h and 12h
    - same as above, but with some dropped values
    - shifted by 3 hours, with duplicates

    Returns
    -------
    ref_frame : pd.DataFrame
        Reference data frame
    test_frames : dict of pd.DataFrame
        Dictionary of data frames, keys are: "shifted_3", "shifted_7",
        "shifted_3_asia", "shifted_7_us", "random_shift", "duplicates".
    expected_nan : dict of np.ndarray
        Dictionary with same keywords as `test_frames`, each entry is a mask
        indicating where NaNs are expected (i.e. no matching was taking place)
    """
    # the reference date range
    ref_dr = pd.date_range("1970", "2020", freq="D", tz="UTC")

    test_dr = {}
    test_dr["shifted_3"] = ref_dr + pd.Timedelta(3, "H")
    test_dr["shifted_7"] = ref_dr + pd.Timedelta(7, "H")
    test_dr["shifted_3_asia"] = test_dr["shifted_3"].tz_convert(
        "Asia/Yekaterinburg"
    )
    test_dr["shifted_7_us"] = test_dr["shifted_7"].tz_convert("US/Eastern")

    # random shifts
    random_hours = np.random.uniform(-12.0, 12.0, len(ref_dr))
    random_mask = np.abs(random_hours) > 6
    dr_random_shift = ref_dr + pd.to_timedelta(random_hours, "H")
    test_dr["random_shift"] = dr_random_shift

    # missing data
    drop_mask = np.zeros(len(ref_dr), dtype=bool)
    drop_mask[100:200] = True
    dr_random_shift = dr_random_shift[~drop_mask]
    test_dr["missing"] = dr_random_shift
    missing_mask = random_mask | drop_mask

    # with duplicates
    test_dr["duplicates"] = deepcopy(test_dr["shifted_3"])
    duplicates_mask = np.zeros(len(ref_dr), dtype=bool)
    for idx in np.random.randint(0, len(test_dr["duplicates"]) - 1, 5):
        test_dr["duplicates"].values[idx] = test_dr["duplicates"].values[
            idx + 1
        ]
        duplicates_mask[idx] = True

    # setting up dataframes
    test_frames = {
        key: pd.DataFrame(
            np.random.randn(len(test_dr[key]), 3), index=test_dr[key]
        )
        for key in test_dr
    }
    ref_frame = pd.DataFrame(np.random.randn(len(ref_dr), 3), index=ref_dr)

    # mask for where we expect nans in the output
    all_nan = np.ones(len(ref_dr), dtype=bool)
    expected_nan = {
        "shifted_3": ~all_nan,
        "shifted_7": all_nan,
        "shifted_3_asia": ~all_nan,
        "shifted_7_us": all_nan,
        "random_shift": random_mask,
        "missing": missing_mask,
        "duplicates": duplicates_mask,
    }
    return ref_frame, test_frames, expected_nan


def setup_data(data, key):
    """Returns only relevant data of test_data for given key"""
    return data[0], data[1][key], data[2][key]


def compare_with_nan(a, b):
    return (a == b) | (np.isnan(a) & np.isnan(b))


def assert_equal_except_nan(res, ref, nan_mask, index_shifted=False):
    expected_nan_idx = nan_mask.nonzero()[0]
    expected_nonan_idx = (~nan_mask).nonzero()[0]
    # using column zero here, all should be the same
    nan_idx = np.isnan(res.iloc[:, 0].values).nonzero()[0]
    nonan_idx = (~np.isnan(res.iloc[:, 0].values)).nonzero()[0]
    assert len(expected_nan_idx) == len(nan_idx)
    if len(nan_idx) > 0:
        assert np.all(nan_idx == expected_nan_idx)
    if len(nonan_idx) > 0 and not index_shifted:
        assert np.all(nonan_idx == expected_nonan_idx)
        assert np.all(
            res.iloc[nonan_idx, 0].values == ref.iloc[nonan_idx, 0].values
        )


@pytest.mark.parametrize(
    "key",
    [
        "shifted_3",
        "shifted_7",
        "shifted_7_us",
        "shifted_3_asia",
        "random_shift",
    ],
)
def test_collocation_nearest_neighbour(test_data, key):
    ref_frame, test_frame, expected_nan = setup_data(test_data, key)
    res = tmatching.temporal_collocation(
        ref_frame, test_frame, pd.Timedelta(6, "H")
    )
    assert_equal_except_nan(res, test_frame, expected_nan)


@pytest.mark.parametrize("key", ["missing", "duplicates"])
def test_collocation_missing_duplicates(test_data, key):
    ref_frame, test_frame, expected_nan = setup_data(test_data, key)
    res = tmatching.temporal_collocation(
        ref_frame,
        test_frame,
        pd.Timedelta(6, "H"),
    )
    # indices of test_frame are shifted w.r.t expected_nan, therefore we can't
    # compare values
    assert_equal_except_nan(res, test_frame, expected_nan, index_shifted=True)


@pytest.mark.parametrize("key", ["shifted_3"])
def test_collocation_window(test_data, key):
    ref_frame, test_frame, expected_nan = setup_data(test_data, key)
    res = tmatching.temporal_collocation(
        ref_frame, test_frame, 6 / 24, dropduplicates=True
    )
    assert_equal_except_nan(res, test_frame, expected_nan, index_shifted=True)


@pytest.mark.parametrize("key", ["shifted_3"])
def test_collocation_input(test_data, key):
    ref_frame, test_frame, expected_nan = setup_data(test_data, key)

    no_timezone = pd.date_range("1970", "2020", freq="D")
    # test with series and index:
    for ref in [ref_frame[0], ref_frame.index, no_timezone]:
        res = tmatching.temporal_collocation(
            ref, test_frame, pd.Timedelta(6, "H")
        )
        assert_equal_except_nan(res, test_frame, expected_nan)


@pytest.mark.parametrize(
    "key",
    [
        "shifted_3",
        "shifted_7",
        "shifted_7_us",
        "shifted_3_asia",
        "random_shift",
    ],
)
def test_collocation_dropna(test_data, key):
    ref_frame, test_frame, expected_nan = setup_data(test_data, key)
    res = tmatching.temporal_collocation(
        ref_frame, test_frame, pd.Timedelta(6, "H"), dropna=True
    )
    expected_nonan_idx = (~expected_nan).nonzero()[0]
    assert np.all(test_frame.iloc[expected_nonan_idx, :].values == res.values)


@pytest.mark.parametrize(
    "key",
    [
        "shifted_3",
        "shifted_7",
        "shifted_7_us",
        "shifted_3_asia",
        "random_shift",
    ],
)
def test_collocation_flag(test_data, key):
    ref_frame, test_frame, expected_nan = setup_data(test_data, key)
    flag = np.random.choice([True, False], len(ref_frame))

    # with array
    res = tmatching.temporal_collocation(
        ref_frame,
        test_frame,
        pd.Timedelta(6, "H"),
        flag=flag,
    )

    compare_with_nan(
        res.iloc[:, 0].values[~flag], test_frame.iloc[:, 0].values[~flag]
    )
    assert np.all(np.isnan(res.values[:, 0][flag]))

    # with array, using invalid as replacement
    res = tmatching.temporal_collocation(
        ref_frame,
        test_frame,
        pd.Timedelta(6, "H"),
        flag=flag,
        use_invalid=True,
    )
    compare_with_nan(res.iloc[:, 0].values, test_frame.iloc[:, 0].values)

    # with dataframe
    test_frame["flag"] = flag
    res = tmatching.temporal_collocation(
        ref_frame,
        test_frame,
        pd.Timedelta(6, "H"),
        flag="flag",
    )
    compare_with_nan(
        res.iloc[:, 0].values[~flag], test_frame.iloc[:, 0].values[~flag]
    )
    assert np.all(np.isnan(res.iloc[:, 0].values[flag]))


# using only shifted_3, because comparison won't work when there are nans
@pytest.mark.parametrize("key", ["shifted_3"])
def test_return_index(test_data, key):
    ref_frame, test_frame, expected_nan = setup_data(test_data, key)
    res = tmatching.temporal_collocation(
        ref_frame, test_frame, pd.Timedelta(6, "H"), return_index=True
    )
    assert_equal_except_nan(res, test_frame, expected_nan)
    assert np.all(test_frame.index.values == res["index_other"].values)


# using only shifted_3, because comparison won't work when there are nans
@pytest.mark.parametrize("key", ["shifted_3", "shifted_7"])
def test_return_distance(test_data, key):
    ref_frame, test_frame, expected_nan = setup_data(test_data, key)
    res = tmatching.temporal_collocation(
        ref_frame, test_frame, pd.Timedelta(6, "H"), return_distance=True
    )
    assert_equal_except_nan(res, test_frame, expected_nan)
    if key == "shifted_3":
        assert np.all(res["distance_other"] == pd.Timedelta(3, "H"))
    if key == "shifted_7":
        assert np.all(np.isnan(res["distance_other"]))


def test_timezone_handling():
    # Issue #150
    data = np.arange(5.0)
    data[3] = np.nan

    match_df = pd.DataFrame(
        {"matched_data": data},
        index=pd.date_range(
            datetime(2007, 1, 1, 0), "2007-01-05", freq="D", tz="UTC"
        ),
    )
    index = pd.DatetimeIndex(
        [
            datetime(2007, 1, 1, 9),
            datetime(2007, 1, 2, 9),
            datetime(2007, 1, 3, 9),
            datetime(2007, 1, 4, 9),
            datetime(2007, 1, 5, 9),
        ]
    ).tz_localize("utc")
    ref_df = pd.DataFrame({"data": np.arange(5)}, index=index)
    matched = tmatching.temporal_collocation(
        ref_df,
        match_df,
        pd.Timedelta(12, "H"),
        dropna=True,
    )

    nptest.assert_allclose(np.array([0, 1, 2, 4]), matched.matched_data)
    assert len(matched) == 4


def test_warning_on_no_match(test_data):
    # Issue #152
    ref_frame, test_frame, expected_nan = setup_data(test_data, "shifted_7")
    with pytest.warns(UserWarning):
        tmatching.temporal_collocation(
            ref_frame, test_frame, pd.Timedelta(6, "H"), checkna=True
        )


def test_combined_matching():
    index = pd.DatetimeIndex([datetime(2007, 1, i + 1, 0) for i in range(10)])
    ref = pd.DatetimeIndex([datetime(2007, 1, i + 1, 5) for i in range(10)])

    data = {
        "data1": np.random.randn(10),
        "data2": np.random.randn(10),
        "missing": np.random.randn(10),
    }
    data["missing"][2] = np.nan
    frames = {
        name: pd.DataFrame({name: data[name]}, index=index) for name in data
    }

    # everything together
    merged = tmatching.combined_temporal_collocation(
        ref,
        (frames[name] for name in frames),
        pd.Timedelta(6, "H"),
        combined_dropna=False,
    )

    assert len(merged) == 10
    for name in frames:
        assert name in merged.columns
        nptest.assert_equal(
            merged[name].values.ravel(), frames[name].values.ravel()
        )

    # test with dropna but not combined_dropna
    merged = tmatching.combined_temporal_collocation(
        ref,
        (frames[name] for name in frames),
        pd.Timedelta(6, "H"),
        combined_dropna=False,
        dropna=True,
    )

    assert len(merged) == 10
    for name in frames:
        assert name in merged.columns
        nptest.assert_equal(
            merged[name].values.ravel(), frames[name].values.ravel()
        )

    # test with combined_dropna
    merged = tmatching.combined_temporal_collocation(
        ref,
        (frames[name] for name in frames),
        pd.Timedelta(6, "H"),
        combined_dropna="any",
        dropna=True,
    )

    assert len(merged) == 9
    for name in frames:
        assert name in merged.columns
        nptest.assert_equal(
            merged[name].values.ravel()[2:], frames[name].values.ravel()[3:]
        )

    # test with 2d-dataframe
    df2d = pd.DataFrame(
        {"2d1": np.random.randn(10), "2d2": np.random.randn(10)}, index=index
    )
    merged = tmatching.combined_temporal_collocation(
        ref,
        (frames["missing"], df2d),
        pd.Timedelta(6, "H"),
        combined_dropna=False,
    )
    assert len(merged) == 10

    # test without match
    for comb_drop in [True, False]:
        merged = tmatching.combined_temporal_collocation(
            ref,
            (frames["missing"], df2d),
            pd.Timedelta(1, "H"),
            combined_dropna=comb_drop,
            dropna=True,
        )
        assert len(merged) == 0


def test_timezone_warning():
    dr = pd.date_range("2000-01-01", "2000-01-31", freq="D")
    dr_berlin = pd.date_range(
        "2000-01-01", "2000-01-31", freq="D", tz="Europe/Berlin"
    )
    n = len(dr)
    with pytest.warns(UserWarning, match="No timezone given"):
        matched = tmatching.temporal_collocation(
            pd.DataFrame(np.random.randn(n), index=dr),
            pd.DataFrame(np.random.randn(n), index=dr_berlin),
            pd.Timedelta(6, "H"),
        )
        assert str(matched.index.tz) == "Europe/Berlin"


def test_combined_timezones():
    dr = pd.date_range("2000-01-01", "2000-01-31", freq="D")
    dr_utc = pd.date_range("2000-01-01", "2000-01-31", freq="D", tz="UTC")
    dr_berlin = pd.date_range(
        "2000-01-01", "2000-01-31", freq="D", tz="Europe/Berlin"
    )
    n = len(dr)

    # test timezone naive
    merged = tmatching.combined_temporal_collocation(
        pd.DataFrame(np.random.randn(n), index=dr),
        (
            pd.DataFrame(np.random.randn(n), index=dr),
            pd.DataFrame(np.random.randn(n), index=dr),
        ),
        pd.Timedelta(6, "H"),
        add_ref_data=True,
    )
    assert merged.index.tz is None

    # test with same timezone
    merged = tmatching.combined_temporal_collocation(
        pd.DataFrame(np.random.randn(n), index=dr_berlin),
        (
            pd.DataFrame(np.random.randn(n), index=dr_berlin),
            pd.DataFrame(np.random.randn(n), index=dr_berlin),
        ),
        pd.Timedelta(6, "H"),
        add_ref_data=True,
    )
    assert str(merged.index.tz) == "Europe/Berlin"

    # test with missing timezone
    with pytest.warns(UserWarning, match="No timezone given"):
        merged = tmatching.combined_temporal_collocation(
            pd.DataFrame(np.random.randn(n), index=dr),
            (
                pd.DataFrame(np.random.randn(n), index=dr_berlin),
                pd.DataFrame(np.random.randn(n), index=dr),
            ),
            pd.Timedelta(6, "H"),
            add_ref_data=True,
        )
        assert str(merged.index.tz) == "Europe/Berlin"

    # test with different timezones and no ref timezone
    with pytest.warns(UserWarning) as warn_record:
        merged = tmatching.combined_temporal_collocation(
            pd.DataFrame(np.random.randn(n), index=dr),
            (
                pd.DataFrame(np.random.randn(n), index=dr_berlin),
                pd.DataFrame(np.random.randn(n), index=dr_utc),
            ),
            pd.Timedelta(6, "H"),
            add_ref_data=True,
        )
        assert str(merged.index.tz) == "UTC"
    assert len(warn_record) == 3
    assert "No timezone given" in warn_record[0].message.args[0]
    assert "Europe/Berlin" in warn_record[0].message.args[0]
    assert "No timezone given" in warn_record[1].message.args[0]
    assert "UTC" in warn_record[1].message.args[0]
    assert "mixed timezones" in warn_record[2].message.args[0]

    # test with different timezones and ref timezone
    merged = tmatching.combined_temporal_collocation(
        pd.DataFrame(np.random.randn(n), index=dr_berlin),
        (
            pd.DataFrame(np.random.randn(n), index=dr_berlin),
            pd.DataFrame(np.random.randn(n), index=dr_utc),
        ),
        pd.Timedelta(6, "H"),
        add_ref_data=True,
    )
    assert str(merged.index.tz) == "Europe/Berlin"


def test_resample_mean():

    times = np.arange(10, dtype=float)
    vals = np.arange(10, dtype=float)

    target_times = np.asarray([3, 7, 10, 12], dtype=float)
    window = 3.0

    expected = np.asarray([3, 7, 9, np.nan])

    resampled = tmatching.resample_mean(times, vals, target_times, window)

    def repeat(x, n):
        return np.repeat(x[np.newaxis, :], n, axis=0)

    newvals = repeat(vals, 12)
    new_expected = repeat(expected, 12)
    resampled = tmatching.resample_mean(times, newvals, target_times, window)
    np.testing.assert_equal(new_expected, resampled)


def test_mean_collocation():
    dr = pd.date_range("1970", "2020", freq="D", tz="UTC")
    other = pd.DataFrame(
        np.vstack(
            (np.arange(len(dr), dtype=float), np.arange(len(dr), dtype=float))
        ).T,
        index=dr,
    )
    # window is actually half window in temporal_collocation
    window = pd.Timedelta(1.5, "D")

    dateidx = list(range(0, len(dr) - 3, 3))
    expected = list(range(0, len(dr) - 3, 3))
    expected[0] = 0.5
    ref_dr = pd.DatetimeIndex([dr[i] for i in dateidx])

    resampled = tmatching.temporal_collocation(
        ref_dr, other, window, method="mean"
    )
    assert (resampled[0] == resampled[1]).all()
    assert (resampled[0] == expected).all()

    # try with Series
    s = tmatching.temporal_collocation(ref_dr, other[0], window, method="mean")
    assert (s == resampled[0]).all()
    assert isinstance(s, pd.Series)


def test_mean_collocation_missing_start_end():
    # In this test the first and last ten timestamps have no values within the
    # window. This is to make sure that indexing at the start and end works as
    # expected.

    dr = pd.date_range("2019", "2020", freq="D", tz="UTC")
    other_dr = pd.date_range("2019", "2020", freq="D", tz="UTC").values
    other_dr[0:10] += pd.Timedelta(12, "H")
    other_dr[-10:] -= pd.Timedelta(12, "H")
    other = pd.DataFrame(
        np.vstack(
            (np.arange(len(dr), dtype=float), np.arange(len(dr), dtype=float))
        ).T,
        index=other_dr,
    )

    resampled = tmatching.temporal_collocation(
        dr, other, pd.Timedelta(6, "H"), method="mean"
    )

    assert np.all(np.isnan(resampled[0:10]))
    assert np.all(np.isnan(resampled[-10:]))
    np.testing.assert_equal(other[10:-10].values, resampled[10:-10].values)

    # test with "random" data that uses a seed with similar properties
    np.random.seed(8986)
    ref = pd.date_range("2020-01-01", "2020-12-31", freq="D")
    values = np.random.randn(len(ref), 3)
    random_shift = np.random.uniform(-12, 12, len(ref))
    random = pd.DataFrame(
        values, index=ref + pd.to_timedelta(random_shift, "H"),
        columns=list(map(lambda x: f"random_{x}", range(3)))
    )
    window = pd.Timedelta(hours=6)
    matched = tmatching.temporal_collocation(
        ref, random, window, method="mean"
    )

    should_be_nan = np.abs(random_shift) > 6
    expected = np.array(values)
    expected[should_be_nan, :] = np.nan
    np.testing.assert_equal(expected, matched)



# Copyright (c) 2016,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology, Department of
#     Geodesy and Geoinformation nor the names of its contributors may be used
#     to endorse or promote products derived from this software without specific
#     prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY, DEPARTMENT OF
# GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

'''
Test for the data manager
'''

import pandas as pd
import pandas.testing as pdtest
import pytest
import numpy as np

import pygeogrids.grids as grids
from pygeobase.io_base import GriddedTsBase

from pytesmo.validation_framework.data_manager import DataManager
from pytesmo.validation_framework.data_manager import get_result_names
from pytesmo.validation_framework.data_manager import get_result_combinations

from tests.test_validation_framework.test_datasets import TestDataset
from tests.test_validation_framework.test_datasets import setup_TestDatasets


class TestDatasetRuntimeError(object):
    """Test dataset that acts as a fake object for the base classes."""
    __test__ = False # prevent pytest from collecting this class during testing

    def __init__(self, filename, mode='r', message="No such file or directory"):
        self.filename = filename
        self.mode = mode
        raise RuntimeError(message)
        open(filename, mode)

    def read(self, gpi):
        return None

    def write(self, gpi, data):
        return None

    def read_ts(self, gpi):
        return None

    def write_ts(self, gpi, data):
        return None

    def close(self):
        pass

    def flush(self):
        pass


def setup_TestDataManager():

    grid = grids.CellGrid(np.array([1, 2, 3, 4]), np.array([1, 2, 3, 4]),
                          np.array([4, 4, 2, 1]), gpis=np.array([1, 2, 3, 4]))

    ds1 = GriddedTsBase("", grid, TestDatasetRuntimeError)
    ds2 = GriddedTsBase("", grid, TestDatasetRuntimeError)
    ds3 = GriddedTsBase("", grid, TestDatasetRuntimeError,
                        ioclass_kws={'message': 'Other RuntimeError'})

    datasets = {
        'DS1': {
            'class': ds1,
            'columns': ['soil moisture'],
            'args': [],
            'kwargs': {}
        },
        'DS2': {
            'class': ds2,
            'columns': ['sm'],
            'args': [],
            'kwargs': {},
            'grids_compatible': True
        },
        'DS3': {
            'class': ds3,
            'columns': ['sm', 'sm2'],
            'args': [],
            'kwargs': {},
            'grids_compatible': True
        }
    }

    dm = DataManager(datasets, 'DS1')
    return dm


def test_DataManager_default_add():

    grid = grids.CellGrid(np.array([1, 2, 3, 4]), np.array([1, 2, 3, 4]),
                          np.array([4, 4, 2, 1]), gpis=np.array([1, 2, 3, 4]))

    ds1 = GriddedTsBase("", grid, TestDataset)

    datasets = {
        'DS1': {
            'class': ds1,
            'columns': ['soil moisture'],
        },
        'DS2': {
            'class': ds1,
            'columns': ['soil moisture'],
        }
    }

    dm = DataManager(datasets, 'DS1')
    assert dm.datasets == {
        'DS1': {
            'class': ds1,
            'columns': ['soil moisture'],
            'args': [],
            'kwargs': {},
            'use_lut': False,
            'max_dist': np.inf,
            'grids_compatible': False
        },
        'DS2': {
            'class': ds1,
            'columns': ['soil moisture'],
            'args': [],
            'kwargs': {},
            'use_lut': False,
            'max_dist': np.inf,
            'grids_compatible': False
        }}


def test_DataManager_read_ts_method_names():

    ds1 = TestDataset("")

    datasets = {
        'DS1': {
            'class': ds1,
            'columns': ['soil moisture'],
        },
        'DS2': {
            'class': ds1,
            'columns': ['soil moisture'],
        }
    }

    read_ts_method_names = {'DS1': 'read_ts',
                            'DS2': 'read_ts_other'}
    dm = DataManager(datasets, 'DS1',
                     read_ts_names=read_ts_method_names)
    data = dm.read_ds('DS1', 1)
    data_other = dm.read_ds('DS2', 1)
    pdtest.assert_frame_equal(data, ds1.read_ts(1))
    pdtest.assert_frame_equal(data_other, ds1.read_ts_other(1))


def test_DataManager_RuntimeError():
    """
    Test DataManager with some fake Datasets that throw RuntimeError
    instead of IOError if a file does not exist like netCDF4

    """

    dm = setup_TestDataManager()
    with pytest.warns(UserWarning):
        dm.read_reference(1)
    with pytest.warns(UserWarning):
        dm.read_other('DS2', 1)
    with pytest.raises(RuntimeError):
        dm.read_other('DS3', 1)


def test_DataManager_dataset_names():

    dm = setup_TestDataManager()
    result_names = dm.get_results_names(3)
    assert result_names == [(('DS1', 'soil moisture'), ('DS2', 'sm'), ('DS3', 'sm')),
                            (('DS1', 'soil moisture'), ('DS2', 'sm'), ('DS3', 'sm2'))]

    result_names = dm.get_results_names(2)
    assert result_names == [(('DS1', 'soil moisture'), ('DS2', 'sm')),
                            (('DS1', 'soil moisture'), ('DS3', 'sm')),
                            (('DS1', 'soil moisture'), ('DS3', 'sm2'))]


def test_DataManager_get_data():

    datasets = setup_TestDatasets()
    dm = DataManager(datasets, 'DS1', read_ts_names={f'DS{i}': 'read' for i in range(1,4)})
    data = dm.get_data(1, 1, 1)
    assert sorted(list(data)) == ['DS1', 'DS2', 'DS3']


def test_get_result_names():

    tst_ds_dict = {'DS1': ['soil moisture'],
                   'DS2': ['sm'],
                   'DS3': ['sm', 'sm2']}
    result_names = get_result_names(tst_ds_dict, 'DS1', 3)
    assert result_names == [(('DS1', 'soil moisture'), ('DS2', 'sm'), ('DS3', 'sm')),
                            (('DS1', 'soil moisture'), ('DS2', 'sm'), ('DS3', 'sm2'))]

    result_names = get_result_names(tst_ds_dict, 'DS1', 2)
    assert result_names == [(('DS1', 'soil moisture'), ('DS2', 'sm')),
                            (('DS1', 'soil moisture'), ('DS3', 'sm')),
                            (('DS1', 'soil moisture'), ('DS3', 'sm2'))]

    result_names = get_result_names(tst_ds_dict, 'DS2', 2)
    assert result_names == [(('DS1', 'soil moisture'), ('DS2', 'sm')),
                            (('DS2', 'sm'), ('DS3', 'sm')),
                            (('DS2', 'sm'), ('DS3', 'sm2'))]

def test_get_result_combinations():

    tst_ds_dict = {'DS1': ['soil moisture'],
                   'DS2': ['sm'],
                   'DS3': ['sm', 'sm2']}
    result_names = get_result_combinations(tst_ds_dict, n=3)
    assert result_names == [(('DS1', 'soil moisture'), ('DS2', 'sm'), ('DS3', 'sm')),
                            (('DS1', 'soil moisture'), ('DS2', 'sm'), ('DS3', 'sm2'))]

    result_names = get_result_combinations(tst_ds_dict, n=2)
    assert result_names == [(('DS1', 'soil moisture'), ('DS2', 'sm')),
                            (('DS1', 'soil moisture'), ('DS3', 'sm')),
                            (('DS1', 'soil moisture'), ('DS3', 'sm2')),
                            (('DS2', 'sm'), ('DS3', 'sm')),
                            (('DS2', 'sm'), ('DS3', 'sm2'))]


def test_maxdist():

    testdf = pd.DataFrame([1, 1, 1], columns=["sm"])
    class TestDataset(object):
        """Test dataset that acts as a fake object for the base classes."""

        def __init__(self, filename, mode='r'):
            self.filename = filename
            self.mode = mode

        def read(self, gpi):
            return testdf

        def read_ts(self, gpi):
            return self.read(gpi)

        def close(self):
            pass

        def flush(self):
            pass

    grid1 = grids.CellGrid(np.array([0, 1]), np.array([0, 1]),
                           np.array([0, 0]))
    grid2 = grids.CellGrid(np.array([0.1, 1]), np.array([0.1, -1]),
                           np.array([0, 0]))

    ds1 = GriddedTsBase("", grid1, TestDataset)
    ds2 = GriddedTsBase("", grid2, TestDataset)

    datasets = {
        'DS1': {
            'class': ds1,
            'columns': ['sm'],
            'args': [],
            'kwargs': {},
            'max_dist': 25e3,  # max dist is in m
        },
        'DS2': {
            'class': ds2,
            'columns': ['sm'],
            'args': [],
            'kwargs': {},
            'max_dist': 25e3,  # max dist is in m
        },
    }

    dm = DataManager(datasets, 'DS1')

    expected = {
        "DS1": testdf,
        "DS2": testdf,
    }

    # test if the close point can be found
    df_dict = dm.get_data(0, 0, 0)
    assert df_dict == expected

    # test if the far away point in the other dataset can be found (should not happen)
    df_dict = dm.get_data(1, 1, 1)
    assert df_dict == {}





# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer thomas.mistelbauer@geo.tuwien.ac.at
# Creation date: 2014-08-13

import unittest
import pytesmo.timedate.dekad as dk
from datetime import datetime


class Test(unittest.TestCase):

    def setUp(self):
        self.begin = datetime(2000, 1, 5)
        self.end = datetime(2000, 3, 15)

        self.date1 = datetime(2000, 2, 1)
        self.date2 = datetime(2000, 2, 13)
        self.date3 = datetime(2000, 2, 28)

    def tearDown(self):
        pass

    def test_dekad_index(self):
        firstdate = datetime(2000, 1, 10)
        lastdate = datetime(2000, 3, 20)
        items = 8

        dkindex = dk.dekad_index(self.begin, self.end)
        assert dkindex[0] == firstdate
        assert dkindex[-1] == lastdate
        assert dkindex.size == items

        dkindex = dk.dekad_index(self.begin)
        assert dkindex[0] == firstdate
        assert (dkindex[-1] - dk.check_dekad(datetime.now())).days == 0

        dkindex = dk.dekad_index(datetime(2014, 1, 3), datetime(2014, 1, 5))
        assert dkindex[0] == datetime(2014, 1, 10)
        assert dkindex.size == 1

        dkindex = dk.dekad_index(datetime(2014, 1, 3), datetime(2014, 1, 15))
        assert dkindex[0] == datetime(2014, 1, 10)
        assert dkindex[1] == datetime(2014, 1, 20)

        dkindex = dk.dekad_index(datetime(2014, 1, 3), datetime(2014, 1, 25))
        assert dkindex[0] == datetime(2014, 1, 10)
        assert dkindex[2] == datetime(2014, 1, 31)

        dkindex = dk.dekad_index(datetime(2014, 1, 13), datetime(2014, 1, 15))
        assert dkindex[0] == datetime(2014, 1, 20)
        assert dkindex.size == 1

        dkindex = dk.dekad_index(datetime(2014, 1, 13), datetime(2014, 1, 25))
        assert dkindex[0] == datetime(2014, 1, 20)
        assert dkindex[1] == datetime(2014, 1, 31)

        dkindex = dk.dekad_index(datetime(2014, 1, 23), datetime(2014, 1, 25))
        assert dkindex[0] == datetime(2014, 1, 31)
        assert dkindex.size == 1

        dkindex = dk.dekad_index(datetime(2014, 1, 3), datetime(2014, 3, 5))
        assert dkindex[0] == datetime(2014, 1, 10)
        assert dkindex[-1] == datetime(2014, 3, 10)

        dkindex = dk.dekad_index(datetime(2014, 1, 13), datetime(2014, 3, 25))
        assert dkindex[0] == datetime(2014, 1, 20)
        assert dkindex[-1] == datetime(2014, 3, 31)

    def test_check_dekad(self):

        dekad1 = dk.check_dekad(self.date1)
        dekad2 = dk.check_dekad(self.date2)
        dekad3 = dk.check_dekad(self.date3)

        assert dekad1 == datetime(2000, 2, 10)
        assert dekad2 == datetime(2000, 2, 20)
        assert dekad3 == datetime(2000, 2, 29)

    def test_dekad2day(self):

        assert dk.dekad2day(2000, 2, 1) == 10
        assert dk.dekad2day(2000, 2, 2) == 20
        assert dk.dekad2day(2000, 2, 3) == 29

    def test_day2dekad(self):

        dekad = dk.day2dekad(29)

        assert dekad == 3

    def test_get_dekad_period(self):

        dates = [self.date1, self.date2, self.date3]

        periods = dk.get_dekad_period(dates)

        assert periods == [4, 5, 6]

    def test_runningdekad2date(self):
        assert dk.runningdekad2date(2014, 35) == datetime(2014, 12, 20)


def test_check_dekad_startdate():
    assert dk.check_dekad_startdate(datetime(2000, 1, 1))
    assert dk.check_dekad_startdate(datetime(2000, 1, 11))
    assert dk.check_dekad_startdate(datetime(2000, 1, 21))
    assert not dk.check_dekad_startdate(datetime(2000, 1, 22))


def test_check_dekad_enddate():
    assert dk.check_dekad_enddate(datetime(2000, 1, 10))
    assert dk.check_dekad_enddate(datetime(2000, 1, 20))
    assert dk.check_dekad_enddate(datetime(2000, 1, 31))
    assert dk.check_dekad_enddate(datetime(2000, 2, 29))
    assert not dk.check_dekad_enddate(datetime(2000, 2, 28))


def test_dekad_startdate_from_date():
    assert datetime(2000, 1, 1) == dk.dekad_startdate_from_date(
        datetime(2000, 1, 10))
    assert datetime(2000, 2, 21) == dk.dekad_startdate_from_date(
        datetime(2000, 2, 29))
    assert datetime(2000, 2, 11) == dk.dekad_startdate_from_date(
        datetime(2000, 2, 17))


def test_group_into_dekads():
    dtimes = [datetime(2000, 1, 10),
              datetime(2000, 1, 11),
              datetime(2000, 1, 12),
              datetime(2000, 1, 23)]
    groups = dk.group_into_dekads(dtimes)
    assert groups == {datetime(2000, 1, 10): [datetime(2000, 1, 10)],
                      datetime(2000, 1, 20): [datetime(2000, 1, 11),
                                              datetime(2000, 1, 12)],
                      datetime(2000, 1, 31): [datetime(2000, 1, 23)]}
    groups = dk.group_into_dekads(dtimes,
                                  use_dekad_startdate=True)
    assert groups == {datetime(2000, 1, 1): [datetime(2000, 1, 10)],
                      datetime(2000, 1, 11): [datetime(2000, 1, 11),
                                              datetime(2000, 1, 12)],
                      datetime(2000, 1, 21): [datetime(2000, 1, 23)]}

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of TU Wien, Department of Geodesy and Geoinformation nor
#     the names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY, DEPARTMENT
# OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from datetime import datetime
import numpy as np
from numpy.testing import (
    assert_equal,
    assert_almost_equal,
    assert_allclose,
)
import pandas as pd
from pathlib import Path
import pytest
import shutil

from pytesmo.metrics import with_analytical_ci, with_bootstrapped_ci, pairwise
from pytesmo.validation_framework.validation import Validation
import pytesmo.validation_framework.error_handling as eh
from pytesmo.validation_framework.metric_calculators import (
    MetadataMetrics,
    BasicMetrics,
    BasicMetricsPlusMSE,
    IntercomparisonMetrics,
    TCMetrics,
    FTMetrics,
    HSAF_Metrics,
    RollingMetrics,
    PairwiseIntercomparisonMetrics,
    TripleCollocationMetrics,
)
from pytesmo.validation_framework.metric_calculators_adapters import (
    MonthsMetricsAdapter,
)

from pytesmo.validation_framework.temporal_matchers import (
    make_combined_temporal_matcher,
    BasicTemporalMatching,
)
from pytesmo.validation_framework.results_manager import netcdf_results_manager
import pytesmo.metrics as metrics


from .utils import DummyReader


def make_some_data():
    """
    Create a data frame with 3 columns and a pre defined bias.
    """
    startdate = datetime(2000, 1, 1)
    enddate = datetime(2000, 12, 31)
    dt_index = pd.date_range(start=startdate, end=enddate, freq="D")

    names = ["ref", "k1", "k2", "k3"]
    # always 0.5
    df = pd.DataFrame(
        index=dt_index,
        data={name: np.repeat(0.5, dt_index.size) for name in names},
    )

    df["k1"] += 0.2  # some positive bias
    df["k2"] -= 0.2  # some negative bias
    df["k3"] -= 0.3  # some more negative bias

    return df


def test_MetadataMetrics_calculator():
    """
    Test MetadataMetrics.
    """
    df = make_some_data()
    data = df[["ref", "k1"]]

    metriccalc = MetadataMetrics(other_name="k1")
    res = metriccalc.calc_metrics(data, gpi_info=(0, 0, 0))

    assert sorted(list(res.keys())) == sorted(["gpi", "lon", "lat", "status"])

    metadata_dict_template = {
        "network": np.array(["None"], dtype="U256"),
        "station": np.array(["None"], dtype="U256"),
        "landcover": np.int32([-1]),
        "climate": np.array(["None"], dtype="U4"),
    }

    metadata_dict = {
        "network": "SOILSCAPE",
        "station": "node1200",
        "landcover": 110,
        "climate": "Csa",
    }

    metriccalc = MetadataMetrics(
        other_name="k1", metadata_template=metadata_dict_template
    )
    res = metriccalc.calc_metrics(data, gpi_info=(0, 0, 0, metadata_dict))
    for key, value in metadata_dict.items():
        assert res[key] == metadata_dict[key]


def test_BasicMetrics_calculator():
    """
    Test BasicMetrics.
    """
    df = make_some_data()
    data = df[["ref", "k1"]]

    metriccalc = BasicMetrics(other_name="k1", calc_tau=False)
    res = metriccalc.calc_metrics(data, gpi_info=(0, 0, 0))

    should = dict(
        n_obs=np.array([366]),
        RMSD=np.array([0.2], dtype="float32"),
        BIAS=np.array([-0.2], dtype="float32"),
    )

    assert res["n_obs"] == should["n_obs"]
    assert np.isnan(res["rho"])
    assert res["RMSD"] == should["RMSD"]
    assert res["BIAS"] == should["BIAS"]
    assert np.isnan(res["R"])

    # scipy 1.3.0 is not built for python 2.7 so we allow both for now
    assert np.isnan(res["p_R"]) or res["p_R"] == 1.0


def test_BasicMetrics_calculator_metadata():
    """
    Test BasicMetrics with metadata.
    """
    df = make_some_data()
    data = df[["ref", "k1"]]

    metadata_dict_template = {"network": np.array(["None"], dtype="U256")}

    metriccalc = BasicMetrics(
        other_name="k1",
        calc_tau=False,
        metadata_template=metadata_dict_template,
    )

    res = metriccalc.calc_metrics(
        data, gpi_info=(0, 0, 0, {"network": "SOILSCAPE"})
    )

    should = dict(
        network=np.array(["SOILSCAPE"], dtype="U256"),
        n_obs=np.array([366]),
        RMSD=np.array([0.2], dtype="float32"),
        BIAS=np.array([-0.2], dtype="float32"),
        dtype="float32",
    )

    assert res["n_obs"] == should["n_obs"]
    assert np.isnan(res["rho"])
    assert res["RMSD"] == should["RMSD"]
    assert res["BIAS"] == should["BIAS"]
    assert res["network"] == should["network"]
    assert np.isnan(res["R"])
    # depends on scipy version changed after v1.2.1
    assert res["p_R"] == np.array([1.0]) or np.isnan(res["R"])


def test_BasicMetricsPlusMSE_calculator():
    """
    Test BasicMetricsPlusMSE.
    """
    df = make_some_data()
    data = df[["ref", "k1"]]

    metriccalc = BasicMetricsPlusMSE(other_name="k1")
    res = metriccalc.calc_metrics(data, gpi_info=(0, 0, 0))

    should = dict(
        network=np.array(["SOILSCAPE"], dtype="U256"),
        n_obs=np.array([366]),
        RMSD=np.array([0.2], dtype="float32"),
        BIAS=np.array([-0.2], dtype="float32"),
        dtype="float32",
    )

    assert res["n_obs"] == should["n_obs"]
    assert np.isnan(res["rho"])
    assert res["RMSD"] == should["RMSD"]
    assert res["BIAS"] == should["BIAS"]
    assert np.isnan(res["R"])
    # depends on scipy version changed after v1.2.1
    assert res["p_R"] == np.array([1.0]) or np.isnan(res["R"])


def test_BasicMetricsPlusMSE_calculator_metadata():
    """
    Test BasicMetricsPlusMSE with metadata.
    """
    df = make_some_data()
    data = df[["ref", "k1"]]

    metadata_dict_template = {"network": np.array(["None"], dtype="U256")}

    metriccalc = BasicMetricsPlusMSE(
        other_name="k1", metadata_template=metadata_dict_template
    )
    res = metriccalc.calc_metrics(
        data, gpi_info=(0, 0, 0, {"network": "SOILSCAPE"})
    )

    should = dict(
        network=np.array(["SOILSCAPE"], dtype="U256"),
        n_obs=np.array([366]),
        RMSD=np.array([0.2], dtype="float32"),
        BIAS=np.array([-0.2], dtype="float32"),
        dtype="float32",
    )

    assert res["n_obs"] == should["n_obs"]
    assert np.isnan(res["rho"])
    assert res["RMSD"] == should["RMSD"]
    assert res["BIAS"] == should["BIAS"]
    assert res["network"] == should["network"]
    assert np.isnan(res["R"])
    # depends on scipy version changed after v1.2.1
    assert res["p_R"] == np.array([1.0]) or np.isnan(res["R"])


def test_IntercompMetrics_calculator():
    """
    Test IntercompMetrics.
    """
    df = make_some_data()
    data = df[["ref", "k1", "k2"]]

    metriccalc = IntercomparisonMetrics(
        other_names=("k1", "k2"), calc_tau=True
    )

    res = metriccalc.calc_metrics(data, gpi_info=(0, 0, 0))

    assert res["n_obs"] == np.array([366])

    assert np.isnan(res["R_between_ref_and_k1"])
    assert np.isnan(res["R_between_ref_and_k2"])

    assert np.isnan(res["rho_between_ref_and_k1"])
    assert np.isnan(res["rho_between_ref_and_k2"])

    np.testing.assert_almost_equal(
        res["mse_between_ref_and_k1"], np.array([0.04], dtype=np.float32)
    )
    np.testing.assert_almost_equal(
        res["mse_between_ref_and_k2"], np.array([0.04], dtype=np.float32)
    )

    np.testing.assert_almost_equal(
        res["mse_corr_between_ref_and_k1"], np.array([0], dtype=np.float32)
    )
    np.testing.assert_almost_equal(
        res["mse_corr_between_ref_and_k2"], np.array([0], dtype=np.float32)
    )

    np.testing.assert_almost_equal(
        res["mse_bias_between_ref_and_k1"], np.array([0.04], dtype=np.float32)
    )
    np.testing.assert_almost_equal(
        res["mse_bias_between_ref_and_k2"], np.array([0.04], dtype=np.float32)
    )

    # scipy 1.3.0 is not built for python 2.7 so we allow both for now
    assert (
        np.isnan(res["p_R_between_ref_and_k1"])
        or res["p_R_between_ref_and_k1"] == 1.0
    )
    assert (
        np.isnan(res["p_R_between_ref_and_k2"])
        or res["p_R_between_ref_and_k2"] == 1.0
    )

    assert res["RMSD_between_ref_and_k1"] == np.array([0.2], dtype="float32")
    assert res["RMSD_between_ref_and_k2"] == np.array([0.2], dtype="float32")

    assert res["BIAS_between_ref_and_k1"] == np.array([-0.2], dtype="float32")
    assert res["BIAS_between_ref_and_k2"] == np.array([0.2], dtype="float32")

    np.testing.assert_almost_equal(
        res["urmsd_between_ref_and_k1"], np.array([0.0], dtype="float32")
    )
    np.testing.assert_almost_equal(
        res["urmsd_between_ref_and_k2"], np.array([0.0], dtype="float32")
    )

    assert "RSS_between_ref_and_k1" in res.keys()
    assert "RSS_between_ref_and_k2" in res.keys()


def test_IntercompMetrics_calculator_metadata():
    """
    Test IntercompMetrics with metadata.
    """
    df = make_some_data()
    data = df[["ref", "k1", "k2"]]

    metadata_dict_template = {"network": np.array(["None"], dtype="U256")}

    metriccalc = IntercomparisonMetrics(
        other_names=("k1", "k2"),
        calc_tau=True,
        metadata_template=metadata_dict_template,
    )
    res = metriccalc.calc_metrics(
        data, gpi_info=(0, 0, 0, {"network": "SOILSCAPE"})
    )

    assert res["network"] == np.array(["SOILSCAPE"], dtype="U256")


def test_TC_metrics_calculator():
    """
    Test TC metrics.
    """
    # this calculator uses a reference data set that is part of ALL triples.
    df = make_some_data()
    data = df[["ref", "k1", "k2", "k3"]]

    metriccalc = TCMetrics(
        other_names=("k1", "k2", "k3"),
        calc_tau=True,
        dataset_names=("ref", "k1", "k2", "k3"),
    )

    res = metriccalc.calc_metrics(data, gpi_info=(0, 0, 0))

    assert res["n_obs"] == np.array([366])

    assert np.isnan(res["R_between_ref_and_k1"])
    assert np.isnan(res["R_between_ref_and_k2"])

    assert np.isnan(res["rho_between_ref_and_k1"])
    assert np.isnan(res["rho_between_ref_and_k2"])

    np.testing.assert_almost_equal(
        res["mse_between_ref_and_k1"], np.array([0.04], dtype=np.float32)
    )
    np.testing.assert_almost_equal(
        res["mse_between_ref_and_k2"], np.array([0.04], dtype=np.float32)
    )

    np.testing.assert_almost_equal(
        res["mse_corr_between_ref_and_k1"], np.array([0], dtype=np.float32)
    )
    np.testing.assert_almost_equal(
        res["mse_corr_between_ref_and_k2"], np.array([0], dtype=np.float32)
    )

    np.testing.assert_almost_equal(
        res["mse_bias_between_ref_and_k1"], np.array([0.04], dtype=np.float32)
    )
    np.testing.assert_almost_equal(
        res["mse_bias_between_ref_and_k2"], np.array([0.04], dtype=np.float32)
    )

    # scipy 1.3.0 is not built for python 2.7 so we allow both for now
    assert (
        np.isnan(res["p_R_between_ref_and_k1"])
        or res["p_R_between_ref_and_k1"] == 1.0
    )
    assert (
        np.isnan(res["p_R_between_ref_and_k2"])
        or res["p_R_between_ref_and_k2"] == 1.0
    )

    assert res["RMSD_between_ref_and_k1"] == np.array([0.2], dtype="float32")
    assert res["RMSD_between_ref_and_k2"] == np.array([0.2], dtype="float32")

    assert res["BIAS_between_ref_and_k1"] == np.array([-0.2], dtype="float32")
    assert res["BIAS_between_ref_and_k2"] == np.array([0.2], dtype="float32")

    np.testing.assert_almost_equal(
        res["urmsd_between_ref_and_k1"], np.array([0.0], dtype="float32")
    )
    np.testing.assert_almost_equal(
        res["urmsd_between_ref_and_k2"], np.array([0.0], dtype="float32")
    )

    assert "RSS_between_ref_and_k1" in res.keys()
    assert "RSS_between_ref_and_k2" in res.keys()
    # each non-ref dataset has a snr, err and beta

    assert np.isnan(res["snr_k1_between_ref_and_k1_and_k2"])
    assert np.isnan(res["snr_k2_between_ref_and_k1_and_k2"])
    assert np.isnan(res["snr_k2_between_ref_and_k2_and_k3"])
    assert np.isnan(res["err_std_k1_between_ref_and_k1_and_k2"])
    np.testing.assert_almost_equal(
        res["beta_k1_between_ref_and_k1_and_k2"][0], 0.0
    )
    np.testing.assert_almost_equal(
        res["beta_k2_between_ref_and_k1_and_k2"][0], 0.0
    )
    np.testing.assert_almost_equal(
        res["beta_k3_between_ref_and_k1_and_k3"][0], 0.0
    )


def test_TC_metrics_calculator_metadata():
    """
    Test TC metrics with metadata.
    """
    df = make_some_data()
    data = df[["ref", "k1", "k2"]]

    metadata_dict_template = {"network": np.array(["None"], dtype="U256")}

    metriccalc = TCMetrics(
        other_names=("k1", "k2"),
        calc_tau=True,
        dataset_names=["ref", "k1", "k2"],
        metadata_template=metadata_dict_template,
    )
    res = metriccalc.calc_metrics(
        data, gpi_info=(0, 0, 0, {"network": "SOILSCAPE"})
    )

    assert res["network"] == np.array(["SOILSCAPE"], dtype="U256")


def test_FTMetrics():
    """
    Test FT metrics.
    """
    df = make_some_data()
    data = df[["ref", "k1"]]

    metriccalc = FTMetrics(frozen_flag=2, other_name="k1")
    res = metriccalc.calc_metrics(data, gpi_info=(0, 0, 0))

    should = dict(
        n_obs=np.array([366]),
        ssf_fr_temp_un=np.array([0.0], dtype="float32"),
        dtype="float32",
    )

    assert res["n_obs"] == should["n_obs"]
    assert res["ssf_fr_temp_un"] == should["ssf_fr_temp_un"]


def test_FTMetrics_metadata():
    """
    Test FT metrics with metadata.
    """
    df = make_some_data()
    data = df[["ref", "k1"]]

    metadata_dict_template = {"network": np.array(["None"], dtype="U256")}

    metriccalc = FTMetrics(
        frozen_flag=2,
        other_name="k1",
        metadata_template=metadata_dict_template,
    )
    res = metriccalc.calc_metrics(
        data, gpi_info=(0, 0, 0, {"network": "SOILSCAPE"})
    )

    assert res["network"] == np.array(["SOILSCAPE"], dtype="U256")


def test_BasicSeasonalMetrics():
    """
    Test BasicSeasonalMetrics.
    """
    df = make_some_data()
    data = df[["ref", "k1"]]

    with pytest.warns(UserWarning):
        metriccalc = MonthsMetricsAdapter(BasicMetrics(other_name="k1"))
        res = metriccalc.calc_metrics(data, gpi_info=(0, 0, 0))

    should = {("ALL", "n_obs"): np.array([366], dtype="float32")}

    assert res[("ALL", "n_obs")] == should[("ALL", "n_obs")]
    assert np.isnan(res[("ALL", "rho")])


def test_BasicSeasonalMetrics_metadata():
    """
    Test BasicSeasonalMetrics with metadata.
    """
    df = make_some_data()
    data = df[["ref", "k1"]]

    metadata_dict_template = {"network": np.array(["None"], dtype="U256")}

    with pytest.warns(UserWarning):
        metriccalc = MonthsMetricsAdapter(
            BasicMetrics(
                other_name="k1", metadata_template=metadata_dict_template
            )
        )
        res = metriccalc.calc_metrics(
            data, gpi_info=(0, 0, 0, {"network": "SOILSCAPE"})
        )

    assert res["network"] == np.array(["SOILSCAPE"], dtype="U256")


def test_HSAF_Metrics():
    """
    Test HSAF Metrics
    """
    df = make_some_data()
    data = df[["ref", "k1", "k2"]]

    metriccalc = HSAF_Metrics(other_name1="k1", other_name2="k2")
    res = metriccalc.calc_metrics(data, gpi_info=(0, 0, 0))

    should = dict(ALL_n_obs=np.array([366]), dtype="float32")

    assert res["ALL_n_obs"] == should["ALL_n_obs"]
    assert np.isnan(res["ref_k1_ALL_rho"])
    assert np.isnan(res["ref_k2_ALL_rho"])


def test_HSAF_Metrics_metadata():
    """
    Test HSAF Metrics with metadata.
    """
    df = make_some_data()
    data = df[["ref", "k1", "k2"]]

    metadata_dict_template = {"network": np.array(["None"], dtype="U256")}

    metriccalc = HSAF_Metrics(
        other_name1="k1", metadata_template=metadata_dict_template
    )
    res = metriccalc.calc_metrics(
        data, gpi_info=(0, 0, 0, {"network": "SOILSCAPE"})
    )

    assert res["network"] == np.array(["SOILSCAPE"], dtype="U256")


def test_RollingMetrics():
    """
    Test RollingMetrics.
    """
    df = make_some_data()
    df["ref"] += np.random.rand(len(df))
    df["k1"] += np.random.rand(len(df))
    data = df[["ref", "k1"]]

    metriccalc = RollingMetrics(other_name="k1")
    dataset = metriccalc.calc_metrics(data, gpi_info=(0, 0, 0), center=False)

    # test pearson r
    ref_array = df["ref"].rolling("30d").corr(df["k1"])
    np.testing.assert_almost_equal(dataset["R"][0], ref_array.values)

    # test rmsd
    indexer = np.arange(30)[None, :] + np.arange(len(df) - 30)[:, None]
    rmsd_arr = []
    for i in range(indexer.shape[0]):
        rmsd_arr.append(
            metrics.rmsd(
                df["ref"][indexer[i, :]].values, df["k1"][indexer[i, :]].values
            )
        )

    rmsd_arr = np.array(rmsd_arr)
    np.testing.assert_almost_equal(dataset["RMSD"][0][29:-1], rmsd_arr)


###########################################################################
# Tests for new QA4SM metrics calculators


# class DummyReader:
#     def __init__(self, df, name):
#         self.data = pd.DataFrame(df[name])

#     def read(self, *args, **kwargs):
#         return self.data


def make_datasets(df):
    datasets = {}
    for key in df:
        ds = {"columns": [key], "class": DummyReader(df, key)}
        datasets[key + "_name"] = ds
    return datasets


def testdata_known_results():
    dr = pd.date_range("2000", "2020", freq="D")
    n = len(dr)
    x = np.ones(n) * 2
    x[10] = np.nan
    df = pd.DataFrame(
        {
            "reference": x,
            # starting with a letter > r here, so we can check how
            # reliable the sorting is
            "plus2": x + 2,
            "minus4": x - 4,
            "plus1": x + 1,
        },
        index=dr,
    )

    expected = {
        (("plus2_name", "plus2"), ("reference_name", "reference")): {
            "n_obs": n - 1,
            "R": np.nan,
            "p_R": np.nan,
            "BIAS": 2,
            "RMSD": 2,
            "mse": 4,
            "RSS": (n - 1) * 4,
            "mse_corr": 0,
            "mse_bias": 4,
            "mse_var": 0,
            "urmsd": 0,
            "gpi": 0,
            "lon": 1,
            "lat": 1,
        },
        (("minus4_name", "minus4"), ("reference_name", "reference")): {
            "n_obs": n - 1,
            "R": np.nan,
            "p_R": np.nan,
            "BIAS": -4,
            "RMSD": 4,
            "mse": 16,
            "RSS": (n - 1) * 16,
            "mse_corr": 0,
            "mse_bias": 16,
            "mse_var": 0,
            "urmsd": 0,
            "gpi": 0,
            "lon": 1,
            "lat": 1,
        },
        (("plus1_name", "plus1"), ("reference_name", "reference")): {
            "n_obs": n - 1,
            "R": np.nan,
            "p_R": np.nan,
            "BIAS": 1,
            "RMSD": 1,
            "mse": 1,
            "RSS": (n - 1) * 1,
            "mse_corr": 0,
            "mse_bias": 1,
            "mse_var": 0,
            "urmsd": 0,
            "gpi": 0,
            "lon": 1,
            "lat": 1,
        },
    }

    # make all arrays of np.float32, except n_obs, gpi (int32) and lat, lon
    # (float64)
    for ck in expected:
        for m in expected[ck]:
            if m in ["n_obs", "gpi"]:
                expected[ck][m] = np.array([expected[ck][m]], dtype=np.int32)
            elif m in ["lat", "lon"]:
                expected[ck][m] = np.array([expected[ck][m]], dtype=np.float64)
            else:
                expected[ck][m] = np.array([expected[ck][m]], dtype=np.float32)

    return make_datasets(df), expected


def testdata_random():
    np.random.seed(42)
    dr = pd.date_range("2000", "2020", freq="D")
    n = len(dr)
    n_datasets = 4

    # generating random correlated data
    r = 0.8
    C = np.ones((n_datasets, n_datasets)) * r
    for i in range(n_datasets):
        C[i, i] = 1
    A = np.linalg.cholesky(C)
    X = (A @ np.random.randn(n_datasets, n)).T

    ref = X[:, 0]
    x1 = X[:, 1]
    x2 = X[:, 2]
    x3 = X[:, 3]
    ref[10] = np.nan
    x2[50] = np.nan
    df = pd.DataFrame(
        {
            "reference": ref,
            "col1": x1,
            "col2": x2,
            "zcol3": x3,
        },
        index=dr,
    )

    expected = {
        (("col1_name", "col1"), ("reference_name", "reference")): {
            "n_obs": n - 2,
            "gpi": 0,
            "lon": 1,
            "lat": 1,
        },
        (("col2_name", "col2"), ("reference_name", "reference")): {
            "n_obs": n - 2,
            "gpi": 0,
            "lon": 1,
            "lat": 1,
        },
        (("reference_name", "reference"), ("zcol3_name", "zcol3")): {
            "n_obs": n - 2,
            "gpi": 0,
            "lon": 1,
            "lat": 1,
        },
    }

    # make all arrays of np.float32, except n_obs, gpi (int32) and lat, lon
    # (float64)
    for ck in expected:
        for m in expected[ck]:
            if m in ["n_obs", "gpi"]:
                expected[ck][m] = np.array([expected[ck][m]], dtype=np.int32)
            elif m in ["lat", "lon"]:
                expected[ck][m] = np.array([expected[ck][m]], dtype=np.float64)
            else:
                expected[ck][m] = np.array([expected[ck][m]], dtype=np.float32)

    return make_datasets(df), expected


@pytest.mark.parametrize(
    "testdata_generator", [testdata_known_results, testdata_random]
)
@pytest.mark.parametrize("seas_metrics", [None, MonthsMetricsAdapter])
def test_PairwiseIntercomparisonMetrics(testdata_generator, seas_metrics):
    # This test first compares the PairwiseIntercomparisonMetrics to known
    # results and then confirms that it agrees with IntercomparisonMetrics as
    # expected

    datasets, expected = testdata_generator()

    # for the pairwise intercomparison metrics it's important that we use
    # make_combined_temporal_matcher

    metrics_calculator = PairwiseIntercomparisonMetrics(
        calc_spearman=True, analytical_cis=False
    )

    if seas_metrics:
        metrics_calculator = seas_metrics(metrics_calculator)

    val = Validation(
        datasets,
        "reference_name",
        scaling=None,  # doesn't work with the constant test data
        temporal_matcher=make_combined_temporal_matcher(pd.Timedelta(6, "H")),
        metrics_calculators={(4, 2): (metrics_calculator.calc_metrics)},
    )
    results_pw = val.calc(
        [0], [1], [1], rename_cols=False, only_with_reference=True
    )

    # in results_pw, there are four entries with keys (("c1name", "c1"),
    # ("refname", "ref"), and so on.
    # Each value is a single dictionary with the values of the metrics

    expected_metrics = [
        "R",
        "p_R",
        "BIAS",
        "RMSD",
        "mse",
        "RSS",
        "mse_corr",
        "mse_bias",
        "urmsd",
        "mse_var",
        "n_obs",
        "gpi",
        "lat",
        "lon",
        "rho",
        "p_rho",
        "tau",
        "p_tau",
    ]
    seasons = ["ALL", "DJF", "MAM", "JJA", "SON"]

    if seas_metrics:
        metrics = []
        for seas in seasons:
            metrics += list(map(lambda x: (seas, x), expected_metrics))
    else:
        metrics = expected_metrics

    for key in results_pw:
        assert isinstance(key, tuple)
        assert len(key) == 2
        assert all(map(lambda x: isinstance(x, tuple), key))
        assert isinstance(results_pw[key], dict)
        res_metrics = list(results_pw[key].keys())
        assert all([v in res_metrics for v in ["lon", "lat", "gpi"]])
        for m in metrics:
            if m in expected[key]:
                assert_equal(results_pw[key][m], expected[key][m])

    # preparation of IntercomparisonMetrics run for comparison
    ds_names = list(datasets.keys())
    metrics = IntercomparisonMetrics(
        dataset_names=ds_names,
        # passing the names here explicitly, see GH issue #220
        refname="reference_name",
        other_names=ds_names[1:],
        calc_tau=True,
    )
    if seas_metrics:
        with pytest.warns(UserWarning):
            metrics = seas_metrics(metrics)
    val = Validation(
        datasets,
        "reference_name",
        scaling=None,
        temporal_matcher=None,  # use default here
        metrics_calculators={(4, 4): metrics.calc_metrics},
    )

    print("running old setup")

    results = val.calc(0, 1, 1, rename_cols=False)

    # results is a dictionary with one entry and key
    # (('c1name', 'c1'), ('c2name', 'c2'), ('c3name', 'c3'), ('refname',
    # 'ref')), the value is a list of length 0, which contains a dictionary
    # with all the results, where the metrics are joined with "_between_" with
    # the combination of datasets, which is joined with "_and_", e.g. for R
    # between ``refname`` and ``c1name`` the key is
    # "R_between_refname_and_c1name"
    if seas_metrics:
        common_metrics = ["gpi", "lat", "lon"]
    else:
        common_metrics = ["n_obs", "gpi", "lat", "lon"]
    pw_metrics = list(set(expected_metrics) - set(common_metrics))
    # there's some sorting done at some point in pytesmo
    oldkey = tuple(sorted([(name, name.split("_")[0]) for name in ds_names]))
    res_old = results[oldkey]

    if seas_metrics:
        metrics = []
        pw_metrics.pop(pw_metrics.index("n_obs"))
        for seas in seasons:
            metrics += list(map(lambda x: (seas, x), pw_metrics))
    else:
        metrics = pw_metrics

    for key in results_pw:
        res = results_pw[key]
        # handle the full dataset metrics
        for m in common_metrics:
            assert_equal(res[m], res_old[m])
        # now get the metrics and compare to the right combination
        for m in metrics:
            othername = key[0][0]
            refname = key[1][0]
            if othername == "reference_name":
                # sorting might be different, see GH #220
                othername = key[1][0]
                refname = key[0][0]
            if isinstance(m, tuple):
                seas, m = m
                old_m_key = (
                    f"{seas}",
                    f"{m}_between_{refname}_and_{othername}",
                )
            else:
                seas = None
                old_m_key = f"{m}_between_{refname}_and_{othername}"
            if m == "BIAS":
                # PairwiseIntercomparisonMetrics has the result as (other,
                # ref), and therefore "bias between other and ref", compared to
                # "bias between ref and bias" in IntercomparisonMetrics
                # this is related to issue #220
                assert_equal(
                    np.abs(res[(seas, m)] if seas else res[m]),
                    np.abs(res_old[old_m_key]),
                )
            elif m == "urmsd":
                # the old implementation differs from the new implementation
                pass
            else:
                assert_equal(
                    res[(seas, m)] if seas else res[m], res_old[old_m_key]
                )


def test_PairwiseIntercomparisonMetrics_confidence_intervals():
    # tests if the correct confidence intervals are returned

    datasets, _ = testdata_random()
    matcher = make_combined_temporal_matcher(pd.Timedelta(6, "H"))
    val = Validation(
        datasets,
        "reference_name",
        scaling=None,  # doesn't work with the constant test data
        temporal_matcher=matcher,
        metrics_calculators={
            (4, 2): (
                PairwiseIntercomparisonMetrics(
                    calc_spearman=True,
                    calc_kendall=True,
                    analytical_cis=True,
                    bootstrap_cis=True,
                ).calc_metrics
            )
        },
    )
    results_pw = val.calc(
        [0], [1], [1], rename_cols=False, only_with_reference=True
    )

    metrics_with_ci = {
        "BIAS": "bias",
        "R": "pearson_r",
        "rho": "spearman_r",
        "tau": "kendall_tau",
    }
    metrics_with_bs_ci = {
        "mse": "msd",
        "mse_bias": "mse_bias",
        "RMSD": "rmsd",
        "urmsd": "ubrmsd",
        "mse_corr": "mse_corr",
        "mse_var": "mse_var",
    }

    # reconstruct dataframe
    frames = []
    for key in datasets:
        frames.append(datasets[key]["class"].data[0])
    data = pd.concat(frames, axis=1)
    data.dropna(how="any", inplace=True)

    for key in results_pw:
        othername = key[0][0]
        other_col = othername.split("_")[0]
        other = data[other_col].values
        refname = key[1][0]
        ref_col = refname.split("_")[0]
        ref = data[ref_col].values
        for metric_key in metrics_with_ci:
            lower = results_pw[key][f"{metric_key}_ci_lower"]
            upper = results_pw[key][f"{metric_key}_ci_upper"]

            # calculate manually from data
            metric_func = getattr(pairwise, metrics_with_ci[metric_key])
            m, lb, ub = with_analytical_ci(metric_func, other, ref)
            # difference due to float32 vs. float64
            assert_almost_equal(upper, ub, 6)
            assert_almost_equal(lower, lb, 6)

        for metric_key in metrics_with_bs_ci:
            lower = results_pw[key][f"{metric_key}_ci_lower"]
            upper = results_pw[key][f"{metric_key}_ci_upper"]

            # calculate manually from data
            metric_func = getattr(pairwise, metrics_with_bs_ci[metric_key])
            m, lb, ub = with_bootstrapped_ci(metric_func, other, ref)
            assert_allclose(upper, ub, rtol=1e-1, atol=1e-4)
            assert_allclose(lower, lb, rtol=1e-1, atol=1e-4)


@pytest.mark.parametrize(
    "testdata_generator", [testdata_known_results, testdata_random]
)
@pytest.mark.parametrize("seas_metrics", [None, MonthsMetricsAdapter])
def test_TripleCollocationMetrics(testdata_generator, seas_metrics):
    # tests by comparison of pairwise metrics to triplet metrics

    datasets, expected = testdata_generator()

    refname = "reference_name"
    othernames = list(datasets.keys())
    othernames.remove(refname)

    triplet_metrics_calculator = TripleCollocationMetrics(
        refname, bootstrap_cis=False
    )

    if seas_metrics:
        triplet_metrics_calculator = seas_metrics(triplet_metrics_calculator)

    matcher = make_combined_temporal_matcher(pd.Timedelta(6, "H"))

    val_triplet = Validation(
        datasets,
        "reference_name",
        scaling=None,  # doesn't work with the constant test data
        temporal_matcher=matcher,
        metrics_calculators={(4, 3): triplet_metrics_calculator.calc_metrics},
    )
    results_triplet = val_triplet.calc(
        [0], [1], [1], rename_cols=False, only_with_reference=True
    )

    if "col1_name" in datasets.keys():
        # we only test the TCA results with the random data, since for the
        # constant data all covariances are zero and TCA therefore doesn't
        # work.
        for metric in ["snr", "err_std", "beta"]:
            for dset in datasets:
                values = []
                dkey = (dset, datasets[dset]["columns"][0])
                for tkey in results_triplet:
                    if dkey in tkey:
                        if seas_metrics is None:
                            values.append(
                                results_triplet[tkey][(metric, dset)][0]
                            )
                        else:
                            values.append(
                                results_triplet[tkey][("DJF", metric, dset)][0]
                            )
                diff = np.abs(np.diff(values))
                assert diff.max() / values[0] < 0.1

    # check if writing to file works
    results_path = Path("__test_results")
    # if this throws, there's either some data left over from previous tests,
    # or some data is named __test_results. Remove the __test_results directory
    # from your current directory to make the test work again.
    assert not results_path.exists()
    results_path.mkdir(exist_ok=True, parents=True)
    netcdf_results_manager(results_triplet, results_path.name)
    assert results_path.exists()
    for key in results_triplet:
        fname = "_with_".join(map(lambda t: ".".join(t), key)) + ".nc"
        assert (results_path / fname).exists()
        # res = xr.open_dataset(results_path / fname)
        # for metric in ["snr", "err_std", "beta"]:
        #     for dset, _ in key:
        #         mkey = metric + "__" + dset
        #         assert mkey in res.data_vars
    shutil.rmtree(results_path)

    # now with CIs, again only for random data
    if "col1_name" in datasets.keys():
        triplet_metrics_calculator = TripleCollocationMetrics(
            refname, bootstrap_cis=True
        )

        if seas_metrics:
            triplet_metrics_calculator = seas_metrics(
                triplet_metrics_calculator
            )

        val_triplet = Validation(
            datasets,
            "reference_name",
            scaling=None,  # doesn't work with the constant test data
            temporal_matcher=matcher,
            metrics_calculators={
                (4, 3): triplet_metrics_calculator.calc_metrics
            },
        )
        results_triplet = val_triplet.calc(
            [0], [1], [1], rename_cols=False, only_with_reference=True
        )
        for key in results_triplet:
            for dset, _ in key:
                for metric in ["snr", "err_std", "beta"]:
                    if seas_metrics is not None:
                        lkey = ("DJF", f"{metric}_ci_lower")
                        ukey = ("DJF", f"{metric}_ci_upper")
                        mkey = ("DJF", f"{metric}")
                    else:
                        lkey = (f"{metric}_ci_lower",)
                        ukey = (f"{metric}_ci_upper",)
                        mkey = (f"{metric}",)
                    assert (*lkey, dset) in results_triplet[key]
                    assert (*ukey, dset) in results_triplet[key]
                    assert (
                        results_triplet[key][(*lkey, dset)]
                        <= results_triplet[key][(*mkey, dset)]
                    )
                    assert (
                        results_triplet[key][(*mkey, dset)]
                        <= results_triplet[key][(*ukey, dset)]
                    )


def test_temporal_matching_ascat_ismn():
    """
    This test uses a CSV file of ASCAT and ISMN data to test if the temporal
    matching within the validation works as epxected in a "real" setup.
    This only tests whether the number of observations matches, because this is
    the main thing the temporal matching influences.
    """

    # test with ASCAT and ISMN data
    here = Path(__file__).resolve().parent
    ascat = pd.read_csv(here / "ASCAT.csv", index_col=0, parse_dates=True)
    ismn = pd.read_csv(here / "ISMN.csv", index_col=0, parse_dates=True)
    dfs = {"ASCAT": ascat, "ISMN": ismn}
    columns = {"ASCAT": "sm", "ISMN": "soil_moisture"}
    refname = "ISMN"
    window = pd.Timedelta(12, "H")

    old_matcher = BasicTemporalMatching().combinatory_matcher
    new_matcher = make_combined_temporal_matcher(window)

    datasets = {}
    for key in ["ISMN", "ASCAT"]:
        all_columns = list(dfs[key].columns)
        ds = {
            "columns": [columns[key]],
            "class": DummyReader(dfs[key], all_columns),
        }
        datasets[key] = ds

    new_val = Validation(
        datasets,
        refname,
        scaling=None,  # doesn't work with the constant test data
        temporal_matcher=new_matcher,
        metrics_calculators={
            (2, 2): PairwiseIntercomparisonMetrics().calc_metrics
        },
    )
    new_results = new_val.calc(
        0, 1, 1, rename_cols=False, only_with_reference=True
    )

    # old setup
    ds_names = list(datasets.keys())
    metrics = IntercomparisonMetrics(
        dataset_names=ds_names,
        # passing the names here explicitly, see GH issue #220
        refname=refname,
        other_names=ds_names[1:],
        calc_tau=True,
    )
    old_val = Validation(
        datasets,
        refname,
        scaling=None,  # doesn't work with the constant test data
        temporal_matcher=old_matcher,
        metrics_calculators={(2, 2): metrics.calc_metrics},
    )
    old_results = old_val.calc(0, 1, 1, rename_cols=False)

    old_key = (("ASCAT", "sm"), ("ISMN", "soil_moisture"))
    new_key = (("ASCAT", "sm"), ("ISMN", "soil_moisture"))

    assert old_results[old_key]["n_obs"] == new_results[new_key]["n_obs"]


def test_TripleCollocationMetrics_failure():
    """
    Test if TripleCollocationMetrics returns the correct status attribute if
    there is not enough data for bootstrapping.
    """
    df = make_some_data()
    data = df.iloc[0:50][["ref", "k1", "k2"]]

    triplet_metrics_calculator = TripleCollocationMetrics(
        "ref", bootstrap_cis=True
    )
    res = triplet_metrics_calculator.calc_metrics(data, gpi_info=(0, 0, 0))
    assert res["status"] == eh.METRICS_CALCULATION_FAILED

# def test_sorting_issue():
# GH #220
# might be a good start for fixing the issue

# dr = pd.date_range("2000", "2020", freq="D")
# n = len(dr)
# x = np.ones(n) * 2
# df = pd.DataFrame(
#     {
#         "reference": x,
#         "zplus2": x + 2,
#         "plus1": x + 1,
#     },
#     index=dr
# )

# datasets = {
#     key + "_name": {"columns": [key], "class": DummyReader(df, key)}
#     for key in df
# }

# val = Validation(
#     datasets,
#     "reference_name",
#     scaling=None,  # doesn't work with the constant test data
#     # temporal_matcher=None,
#     temporal_matcher=make_combined_temporal_matcher(
#          pd.Timedelta(6, "H")
#     ),
#     metrics_calculators={
#         (3, 2): PairwiseIntercomparisonMetrics().calc_metrics
#     }
# )
# results = val.calc(1, 1, 1)

# assert 0
